#include "Plane3.h"
#include "ViewCellBsp.h"
#include "Mesh.h"
#include "common.h"
#include "ViewCell.h"
#include "Environment.h"
#include "Polygon3.h"
#include "Ray.h"
#include "AxisAlignedBox3.h"
#include <stack>
#include <time.h>
#include <iomanip>
#include "Exporter.h"

#define INITIAL_COST 999999// unreachable high initial cost for heuristic evaluation

int BspTree::sTermMaxPolygons = 10;
int BspTree::sTermMaxDepth = 20;
int BspTree::sMaxCandidates = 10;
int BspTree::sSplitPlaneStrategy = NEXT_POLYGON; 
int BspTree::sConstructionMethod = VIEW_CELLS;

/** Evaluates split plane classification with respect to the plane's 
	contribution for a balanced tree.
*/
int BspTree::sLeastSplitsTable[4] = {0, 0, 1, 0};
/** Evaluates split plane classification with respect to the plane's 
	contribution for a minimum number splits in the tree.
*/
int BspTree::sBalancedTreeTable[4] = {-1, 1, 0, 0};

const int FACTOR_BALANCE = 1;
const int FACTOR_LEAST_SPLITS = 1;

int counter = 0;
bool BspTree::displayDebug = false;
/****************************************************************/
/*                  class BspNode implementation                */
/****************************************************************/

BspNode::BspNode(): mParent(NULL), mPolygons(NULL),mViewCellIdx(0)
{}

BspNode::BspNode(BspInterior *parent): mParent(parent), mPolygons(NULL),mViewCellIdx(0)
{}

BspNode::~BspNode()
{
	if (mPolygons)
	{
		CLEAR_CONTAINER(*mPolygons);
		DEL_PTR(mPolygons);
	}
}

bool BspNode::IsRoot() const 
{
	return mParent == NULL;
}

BspInterior *BspNode::GetParent() 
{ 
	return mParent; 
}

void BspNode::SetParent(BspInterior *parent)
{
	mParent = parent;
}

PolygonContainer *BspNode::GetPolygons()
{
	if (!mPolygons)
		mPolygons = new PolygonContainer();

	return mPolygons;
}

void BspNode::ProcessPolygons(PolygonContainer *polys, bool storePolys)
{
	if (storePolys)
	{
		while (!polys->empty())
		{
			GetPolygons()->push_back(polys->back());
			polys->pop_back();
		}
	}
	else CLEAR_CONTAINER(*polys);

	delete polys;
}


/****************************************************************/
/*              class BspInterior implementation                */
/****************************************************************/


BspInterior::BspInterior(const Plane3 &plane): 
mPlane(plane), mFront(NULL), mBack(NULL)
{}

bool BspInterior::IsLeaf() const
{ 
	return false; 
}

BspNode *BspInterior::GetBack() 
{
	return mBack;
}

BspNode *BspInterior::GetFront() 
{
	return mFront;
}

Plane3 *BspInterior::GetPlane()
{
	return &mPlane;
}

void BspInterior::ReplaceChildLink(BspNode *oldChild, BspNode *newChild) 
{
	if (mBack == oldChild)
		mBack = newChild;
	else
		mFront = newChild;
}

void BspInterior::SetupChildLinks(BspNode *b, BspNode *f) 
{
    mBack = b;
    mFront = f;
}

void BspInterior::ProcessPolygon(Polygon3 *poly, const bool storePolys)
{
	if (storePolys) 
		GetPolygons()->push_back(poly);
	else
		delete poly;
}

bool BspInterior::SplitPolygons(PolygonContainer *polys, 
								PolygonContainer *frontPolys, 
								PolygonContainer *backPolys, 
								PolygonContainer *coincident, 
								int &splits, bool storePolys)
{
#ifdef _Debug
	if (BspTree::displayDebug)Debug << "Splitting polygons of node " << this << " with plane " << mPlane << endl;
#endif
	while (!polys->empty())
	{
		Polygon3 *poly = polys->back();
		polys->pop_back();

		//if (BspTree::displayDebug) Debug << "New polygon with plane: " << poly->GetSupportingPlane() << "\n";

		// test if split is neccessary
		int classification = poly->ClassifyPlane(mPlane);

		Polygon3 *front_piece = NULL;
		Polygon3 *back_piece = NULL;
	
		switch (classification)
		{
			case Polygon3::COINCIDENT:
				// same surface normal
				if (DotProd(mPlane.mNormal, poly->GetSupportingPlane().mNormal) > 0)
				{
					coincident->push_back(poly);
					break;
				}
				//if (BspTree::displayDebug)	Debug << "coincident" << endl;				
				
			case Polygon3::FRONT_SIDE:	
				//if (BspTree::displayDebug)Debug << "front" << endl;
				frontPolys->push_back(poly);
				break;
			case Polygon3::BACK_SIDE:
				//if (BspTree::displayDebug)Debug << "back" << endl;
				backPolys->push_back(poly);
				break;
			case Polygon3::SPLIT:
				front_piece = new Polygon3(poly->GetParent());
				back_piece = new Polygon3(poly->GetParent());

				//-- split polygon
				poly->Split(&mPlane, front_piece, back_piece, splits);

				frontPolys->push_back(front_piece);
				backPolys->push_back(back_piece);

#ifdef _DEBUG
				if (BspTree::displayDebug)Debug << "split " << *poly << endl << *front_piece << endl << *back_piece << endl;
#endif
				ProcessPolygon(poly, storePolys);
				
				break;
			default:
                Debug << "SHOULD NEVER COME HERE\n";
				break;
		}
	}
	//if (BspTree::displayDebug) Debug << "inside: " << inside << endl;
	// contains nothing
	delete polys;
}

/****************************************************************/
/*                  class BspLeaf implementation                */
/****************************************************************/
BspLeaf::BspLeaf(): mViewCell(NULL)
{
}

BspLeaf::BspLeaf(ViewCell *viewCell): mViewCell(viewCell)
{
}

BspLeaf::BspLeaf(BspInterior *parent): BspNode(parent), mViewCell(NULL)
{}

BspLeaf::BspLeaf(BspInterior *parent, ViewCell *viewCell): 
BspNode(parent), mViewCell(viewCell)
{
}
ViewCell *BspLeaf::GetViewCell()
{
	return mViewCell;
}

void BspLeaf::SetViewCell(ViewCell *viewCell)
{
	mViewCell = viewCell;
}

bool BspLeaf::IsLeaf() const 
{ 
	return true; 
}

/****************************************************************/
/*                  class BspTree implementation                */
/****************************************************************/

BspTree::BspTree(): 
mTermMaxPolygons(0), 
mTermMaxDepth(0), 
mRoot(NULL), 
mIsIncremential(false),
//mStorePolys(true)
mStorePolys(false)
{
	Randomize(); // initialise random generator for heuristics
}
 
const BspTreeStatistics &BspTree::GetStatistics() const 
{
	return mStat;
}

void BspTreeStatistics::Print(ostream &app) const
{
	app << "===== BspTree statistics ===============\n";

	app << setprecision(4);

	app << "#N_RAYS Number of rays )\n"
		<< rays << endl;
	app << "#N_DOMAINS  ( Number of query domains )\n"
		<< queryDomains <<endl;

	app << "#N_NODES ( Number of nodes )\n" << nodes << "\n";

	app << "#N_INTERIORS ( Number of interior nodes )\n" << Interior() << "\n";

	app << "#N_LEAVES ( Number of leaves )\n" << Leaves() << "\n";

	app << "#N_SPLITS ( Number of splits )\n" << splits << "\n";

	app << "#N_RAYREFS ( Number of rayRefs )\n" <<
	rayRefs << "\n";

	app << "#N_RAYRAYREFS  ( Number of rayRefs / ray )\n" <<
	rayRefs/(double)rays << "\n";

	app << "#N_LEAFRAYREFS  ( Number of rayRefs / leaf )\n" <<
	rayRefs/(double)Leaves() << "\n";

	app << "#N_MAXOBJECTREFS  ( Max number of rayRefs / leaf )\n" <<
	maxObjectRefs << "\n";

	app << "#N_NONEMPTYRAYREFS  ( Number of rayRefs in nonEmpty leaves / non empty leaf )\n" <<
	rayRefsNonZeroQuery/(double)(Leaves() - zeroQueryNodes) << "\n";

	app << "#N_LEAFDOMAINREFS  ( Number of query domain Refs / leaf )\n" <<
	objectRefs/(double)Leaves() << "\n";

	app << "#N_PEMPTYLEAVES  ( Percentage of leaves with zero query domains )\n"<<
	zeroQueryNodes*100/(double)Leaves() << endl;

	app << "#N_PMAXDEPTHLEAVES ( Percentage of leaves at maxdepth )\n"<<
	maxDepthNodes * 100 / (double)Leaves() << endl;

	app << "#N_PMAXDEPTH ( Maximal reached depth )\n" << maxDepth << endl;

	app << "#N_PMINDEPTH ( Minimal reached depth )\n" << minDepth << endl;

	app << "#AVGDEPTH ( average depth )\n" << AvgDepth() << endl;

	app << "#N_ADDED_RAYREFS  ( Number of dynamically added ray references )\n"<<
	addedRayRefs << endl;

	app << "#N_REMOVED_RAYREFS  ( Number of dynamically removed ray references )\n" <<
	removedRayRefs << endl;

	app << "#N_INPUT_POLYGONS (number of input polygons )\n" <<
		polys << endl;

	app << "#N_ROUTPUT_INPUT_POLYGONS ( ratio polygons after subdivision / input polygons )\n" <<
		 (polys + splits) / (double)polys << endl;
	
	app << "===== END OF BspTree statistics ==========\n";
}

BspTree::~BspTree()
{
	std::stack<BspNode *> tStack;

	tStack.push(mRoot);

	while (!tStack.empty())
	{
		BspNode *node = tStack.top();

	    tStack.pop();
	
		if (!node->IsLeaf())
		{
			BspInterior *interior = dynamic_cast<BspInterior *>(node);

			// push the children on the stack (there are always two children)
			tStack.push(interior->GetBack());
			tStack.push(interior->GetFront());
		}

		DEL_PTR(node);
	}
}


void BspTree::InsertViewCell(ViewCell *viewCell)
{	
	std::stack<BspTraversalData> tStack;
	
	PolygonContainer *polys = new PolygonContainer();

	// extract polygons that guide the split process
	mStat.polys += AddMesh2Polygons(viewCell->GetMesh(), *polys);
	mBox.Include(viewCell->GetBox()); // add to BSP aabb

	 // traverse tree or create new one
	BspNode *firstNode = mRoot ? mRoot : new BspLeaf();

	tStack.push(BspTraversalData(firstNode, polys, 0, true));

	while (!tStack.empty())
	{
		// filter polygons donw the tree
		BspTraversalData tData = tStack.top();
	    tStack.pop();
			
		if (!tData.mNode->IsLeaf())
		{
			BspInterior *interior = dynamic_cast<BspInterior *>(tData.mNode);

			//-- filter view cell polygons down the tree until a leaf is reached
			if ((int)tData.mPolygons->size() > 0)
			{
				PolygonContainer *frontPolys = new PolygonContainer();
				PolygonContainer *backPolys = new PolygonContainer();
				PolygonContainer *coincident = new PolygonContainer();

				int splits = 0;
		
				// split viecell polygons with respect to split plane
				bool inside = interior->SplitPolygons(tData.mPolygons, 
													  frontPolys, 
													  backPolys,
													  coincident,
													  splits, 
													  mStorePolys);
				
				//Debug << "split node on level " << tData.mDepth << " bk: " << backPolys->size() << " frnt: " << frontPolys->size() << endl;
				mStat.splits += splits;

				// push the children on the stack
				tStack.push(BspTraversalData(interior->GetFront(), frontPolys, tData.mDepth + 1));
				tStack.push(BspTraversalData(interior->GetBack(), backPolys, tData.mDepth + 1));
			
			}
			else // stop traversal
			{
				DEL_PTR(tData.mPolygons);
			}
		}
		else // reached leaf => subdivide current viewcell
		{
			if (tData.mPolygons->size() > 0)
				Debug << "WARNING (should not come here): polygon size: " << (int)tData.mPolygons->size() << " inside: " << tData.mIsInside  << endl;

			BspNode *root = Subdivide(tStack, tData, viewCell);		

			if (!mRoot) // tree empty => new root
				mRoot = root;
		}
	}
}

void BspTree::InitTree(int maxPolygons, int maxDepth)
{
	mTermMaxPolygons = maxPolygons;
	mTermMaxDepth = maxDepth;
	mStat.nodes = 1;
	
	mBox.Initialize(); 	// initialise bsp tree bounding box
}



int BspTree::AddMesh2Polygons(Mesh *mesh, PolygonContainer &polys)
{
	FaceContainer::const_iterator fi;
	int polysNum = 0;

	// copy the face data to polygons
	for (fi = mesh->mFaces.begin(); fi != mesh->mFaces.end(); ++ fi)
	{
		Polygon3 *poly = new Polygon3((*fi), mesh);
		polys.push_back(poly);
		//if (displayDebug)Debug << *poly << endl;
		polysNum ++;
	}
	return polysNum;
}

int BspTree::Copy2PolygonSoup(const ViewCellContainer &viewCells, PolygonContainer &polys, int maxObjects)
{
	int limit = (maxObjects > 0) ? Min((int)viewCells.size(), maxObjects) : (int)viewCells.size();
  
	for (int i = 0; i < limit; ++i)
	{//if ((i == 12) ||(i==14))
		if (viewCells[i]->GetMesh()) // copy the mesh data to polygons
		{
			mBox.Include(viewCells[i]->GetBox()); // add to BSP tree aabb
			AddMesh2Polygons(viewCells[i]->GetMesh(), polys);
		}
	}

	return (int)polys.size();
}

int BspTree::Copy2PolygonSoup(const ObjectContainer &objects, PolygonContainer &polys, int maxObjects)
{
	int limit = (maxObjects > 0) ? Min((int)objects.size(), maxObjects) : (int)objects.size();
  
	for (int i = 0; i < limit; ++i)
	{
		Intersectable *object = objects[i];//*it;
		Mesh *mesh = NULL;

		switch (object->Type()) // extract the meshes
		{
		case Intersectable::MESH_INSTANCE:
			mesh = dynamic_cast<MeshInstance *>(object)->GetMesh();
			break;
		case Intersectable::VIEWCELL:
			mesh = dynamic_cast<ViewCell *>(object)->GetMesh();
			break;
			// TODO: handle transformed mesh instances
		default:
			break;
		}
		
        if (mesh) // copy the mesh data to polygons
		{
			mBox.Include(object->GetBox()); // add to BSP tree aabb
			AddMesh2Polygons(mesh, polys);
		}
	}

	return (int)polys.size();
}

void BspTree::Construct(const ViewCellContainer &viewCells)
{
	//-- Construct tree using the given viewcells
	
	/* for this type of construction we filter all viewcells down the 
	 * tree. If there is no polygon left, the last split plane
	 * decides inside or outside of the viewcell.
    */
	InitTree(sTermMaxPolygons, sTermMaxDepth);

    bool savedStorePolys = mStorePolys;

	// tree is completely constructed once before 
	// view cells are inserted one after another => 
	// global tree optimization possible
	if (!mIsIncremential) // todo: not incremential does not work
	{
		// copy view cell meshes into one big polygon soup
		PolygonContainer *polys = new PolygonContainer();
		Copy2PolygonSoup(viewCells, *polys);

		// polygons are stored only during view cell insertion
		mStorePolys = false;

		// construct tree from viewcell polygons
		Construct(polys);
		//Export("bsp.x3d");
	}
	
	//-- insert all viewcells one after another
	ViewCellContainer::const_iterator it;

	mStorePolys = savedStorePolys;
	mStat.polys = 0;
	mStat.splits = 0;
	mStat.accumDepth = 0;

	long startTime = GetTime();
//displayDebug = true;
	counter = 0;
	Debug << "**** Starting view cell insertion ****" << endl;
	for (it = viewCells.begin(); it != viewCells.end(); ++ it)
	{
		//if ((counter == 12) || (counter == 14)){
		//Debug << "** inserting view cell " << counter << " **" << endl;
		InsertViewCell(*it);
	//}	counter ++;
	}

	Debug << "**** Finished view cell insertion ****" << endl;
	Debug << "insertion time: "<< TimeDiff(startTime, GetTime())*1e-3 << "s" << endl;
}


void BspTree::Construct(const ObjectContainer &objects, ViewCellContainer *viewCells)
{
	// take termination criteria from globals
	InitTree(sTermMaxPolygons, sTermMaxDepth);
	
	PolygonContainer *polys = new PolygonContainer();
	
	// copy mesh instance polygons into one big polygon soup
	mStat.polys = Copy2PolygonSoup(objects, *polys);

	// construct tree from polygon soup
	Construct(polys, viewCells);
}

void BspTree::Construct(const RayContainer &rays, ViewCellContainer *viewCells)
{
	// TODO
}

void BspTree::Construct(PolygonContainer *polys, ViewCellContainer *viewCells)
{
	std::stack<BspTraversalData> tStack;
	
	BspTraversalData tData(new BspLeaf(), polys, 0, true, viewCells);

	tStack.push(tData);

	long startTime = GetTime();
	Debug << "**** Contructing tree using objects ****\n";
	while (!tStack.empty()) 
	{
		tData = tStack.top();
	    tStack.pop();

		//TODO
		ViewCell *viewCell = NULL;

        /*if (viewCells) // generate new view cell in leaf
		{
			viewCell = new ViewCell();
			viewCells->push_back(viewCell);
		}*/

		// subdivide leaf node
		BspNode *root = Subdivide(tStack, tData, viewCell);

		// empty tree => new root corresponding to unbounded space
		if (!mRoot) 
			mRoot = root;
	}

	Debug << "**** Finished tree construction ****\n";
	Debug << "BSP tree contruction time: "<< TimeDiff(startTime, GetTime())*1e-3 << "s" << endl;
}


BspNode *BspTree::Subdivide(BspTraversalStack &tStack, BspTraversalData &tData, 
							ViewCell *viewCell)
{
	//-- terminate traversal  
	if ((tData.mPolygons->size() <= mTermMaxPolygons) || (tData.mDepth >= mTermMaxDepth))
		// if there is another view cell associated with this leaf => subdivide further
		//&& !(viewCell && tData.mIsInside && dynamic_cast<BspLeaf *>(tData.mNode)->GetViewCell()))  
	{
#ifdef _DEBUG
		if (displayDebug)Debug << "subdivision terminated at depth " << tData.mDepth << ", #polys: " << (int)tData.mPolygons->size() << endl;
#endif

		EvaluateLeafStats(tData);

	    // add view cell if inside object) 
		if (viewCell && tData.mCoincident && (tData.mCoincident->size() > 0))) // || (tData.mGeometry->size() > 0) // or if there is still geometry left 
		{
			BspLeaf *leaf = dynamic_cast<BspLeaf *>(tData.mNode);

			if (leaf->GetViewCell())
				Debug << "ERROR: leaf already has view cell " << endl;//leaf->mViewCellIdx << endl;

			//leaf->mViewCellIdx = counter;
			//Debug << "insert view cell" << endl;

			leaf->SetViewCell(coincident[0]->GetParent());
		}

		// add or delete remaining polygons
		tData.mNode->ProcessPolygons(tData.mPolygons, mStorePolys);
		tData.mNode->ProcessPolygons(tData.coincident, mStorePolys);

		return tData.mNode;
	}

	tData.mNode->ProcessPolygons(tData.coincident, mStorePolys);

	//-- create new subdivided node
	PolygonContainer *backPolys = new PolygonContainer();
	PolygonContainer *frontPolys = new PolygonContainer();
	PolygonContainer *coincident = new PolygonContainer();
	
	BspInterior *interior = SubdivideNode(dynamic_cast<BspLeaf *>(tData.mNode),
										  tData.mPolygons,
										  frontPolys,
										  backPolys, 
										  coincident);

	// push the children on the stack
	// inside information is only propagated with the back leaf
	tStack.push(BspTraversalData(interior->GetBack(), backPolys, tData.mDepth + 1, coincident));
	tStack.push(BspTraversalData(interior->GetFront(), frontPolys, tData.mDepth + 1));

	return interior;
}


BspInterior *BspTree::SubdivideNode(BspLeaf *leaf, 
									PolygonContainer *polys, 
									PolygonContainer *frontPolys,
									PolygonContainer *coincident, 
									PolygonContainer *backPolys, bool &inside)
{
	mStat.nodes += 2;

	// add the new nodes to the tree + select subdivision plane
	BspInterior *interior = new BspInterior(SelectPlane(polys)); 

#ifdef _DEBUG
	Debug << interior << endl;
#endif

	// split polygon according to current plane
	int splits = 0;
		
	inside = interior->SplitPolygons(polys, frontPolys, backPolys, coincident, splits, mStorePolys);
	
 	mStat.splits += splits;

	BspInterior *parent = leaf->GetParent();

	// replace a link from node's parent
	if (parent)
	{
		parent->ReplaceChildLink(leaf, interior);
		interior->SetParent(parent);
	}

	// and setup child links
	//interior->SetupChildLinks(new BspLeaf(interior), new BspLeaf(interior, leaf->mViewCell));
	interior->SetupChildLinks(new BspLeaf(interior), new BspLeaf(interior));
	
	delete leaf; // leaf not member of tree anymore

	return interior;
}

Plane3 BspTree::SelectPlane(PolygonContainer *polys)  const
{
	if (polys->size() == 0)
	{
		Debug << "Warning: No split candidate available\n";
		return Plane3();
	}

	// simple strategy: just take next polygon
	if (sSplitPlaneStrategy == NEXT_POLYGON)
	{
		return polys->front()->GetSupportingPlane();
	}
	
	// use heuristics to find appropriate plane
	return SelectPlaneHeuristics(polys, sMaxCandidates);
}

Plane3 BspTree::SelectPlaneHeuristics(PolygonContainer *polygons, int maxTests) const
{
	int lowestCost = INITIAL_COST;
	int bestPlaneIdx = 0;
	
	int limit = Min((int)polygons->size(), maxTests);

	for (int i = 0; i < limit; ++ i)
	{
		int candidateIdx = Random((int)polygons->size()); // TODO:  avoid testing same plane 2 times
		Plane3 candidatePlane = (*polygons)[candidateIdx]->GetSupportingPlane();
		
		// evaluate current candidate
		int candidateCost = EvalSplitPlane(polygons, candidatePlane);
			
		if (candidateCost < lowestCost)
		{
			bestPlaneIdx = candidateIdx;
			lowestCost = candidateCost;
			//Debug << "new plane cost " << lowestCost << endl;
		}
	}
	//Debug << "Plane lowest cost: " << lowestCost << endl;
	
	return (*polygons)[bestPlaneIdx]->GetSupportingPlane();
}

int BspTree::EvalSplitPlane(PolygonContainer *polygons, const Plane3 &candidatePlane) const
{
	PolygonContainer::const_iterator it;

	int sumBalanced = 0, sumLeastSplits = 0;

	for (it = polygons->begin(); it != polygons->end(); ++ it)
	{
		int classification = (*it)->ClassifyPlane(candidatePlane);
		
		if ((sSplitPlaneStrategy == COMBINED) || (sSplitPlaneStrategy == BALANCED_TREE))
		{
			sumBalanced += sBalancedTreeTable[classification];
		}
		
		if ((sSplitPlaneStrategy == COMBINED) || (sSplitPlaneStrategy == LEAST_SPLITS))
		{
			sumLeastSplits += sLeastSplitsTable[classification];
		}
	}

	// return linear combination of both sums
	return FACTOR_BALANCE * abs(sumBalanced) + FACTOR_LEAST_SPLITS *abs(sumLeastSplits);
}

void BspTree::ParseEnvironment()
{
	environment->GetIntValue("BspTree.Termination.maxDepth", sTermMaxDepth);
	environment->GetIntValue("BspTree.Termination.maxPolygons", sTermMaxPolygons);
	environment->GetIntValue("BspTree.maxCandidates", sMaxCandidates);

	Debug << "BSP max depth: " << sTermMaxDepth << endl;
	Debug << "BSP max polys: " << sTermMaxPolygons << endl;
	Debug << "BSP max candidates: " << sMaxCandidates << endl;

	//-- extract strategy to choose the next split plane
	char splitPlaneStrategyStr[60];

	environment->GetStringValue("BspTree.splitPlaneStrategy", splitPlaneStrategyStr);
	
	sSplitPlaneStrategy = BspTree::NEXT_POLYGON;
	
	if (strcmp(splitPlaneStrategyStr, "nextPolygon") == 0)
		sSplitPlaneStrategy = BspTree::NEXT_POLYGON;
	else if (strcmp(splitPlaneStrategyStr, "leastSplits") == 0)
		sSplitPlaneStrategy = BspTree::LEAST_SPLITS;
	else if (strcmp(splitPlaneStrategyStr, "balancedTree") == 0)
		sSplitPlaneStrategy = BspTree::BALANCED_TREE;
	else if (strcmp(splitPlaneStrategyStr, "combined") == 0)
		sSplitPlaneStrategy = BspTree::COMBINED;
	else 
	{
		cerr << "Wrong BSP split plane strategy " << splitPlaneStrategyStr << endl;
		exit(1);
    }

	Debug << "Split plane heuristic: " << splitPlaneStrategyStr << endl;

	//-- parse BSP tree construction method
	char constructionMethodStr[60];
	
	environment->GetStringValue("BspTree.constructionMethod", constructionMethodStr);

	sConstructionMethod = BspTree::VIEW_CELLS;
	
	if (strcmp(constructionMethodStr, "viewCells") == 0)
		sConstructionMethod = BspTree::VIEW_CELLS;
	else if (strcmp(constructionMethodStr, "sceneGeometry") == 0)
		sConstructionMethod = BspTree::SCENE_GEOMETRY;
	else if (strcmp(constructionMethodStr, "rays") == 0)
		sConstructionMethod = BspTree::RAYS;
	else 
	{
		cerr << "Wrong bsp construction method " << constructionMethodStr << endl;
		exit(1);
    }

	Debug << "Construction method: " << constructionMethodStr << endl;
}

void BspTree::CollectLeaves(vector<BspLeaf *> &leaves)
{
	stack<BspNode *> nodeStack;
	nodeStack.push(mRoot);
 
	while (!nodeStack.empty()) 
	{
		BspNode *node = nodeStack.top();
    
		nodeStack.pop();
    
		if (node->IsLeaf()) 
		{
			BspLeaf *leaf = (BspLeaf *)node;
			
			leaves.push_back(leaf);
		} else 
		{
			BspInterior *interior = dynamic_cast<BspInterior *>(node);
			nodeStack.push(interior->GetBack());
			nodeStack.push(interior->GetFront());
		}
	}
}

int BspTree::CollectLeafPvs()
{
	int totalPvsSize = 0;
  
	stack<BspNode *> nodeStack;
  
	nodeStack.push(mRoot);
  
	while (!nodeStack.empty()) 
	{
		BspNode *node = nodeStack.top();
		
		nodeStack.pop();
		
		if (node->IsLeaf()) 
		{
			BspLeaf *leaf = dynamic_cast<BspLeaf *>(node);

			ViewCell *viewcell = leaf->GetViewCell();
			
			if (!viewcell->Mailed()) 
			{
				viewcell->Mail(); // what does mail mean?
					
				// add this node to pvs of all nodes it can see
				BspPvsMap::iterator ni; 
	  
	/*			for (ni = object->mBspPvs.mEntries.begin(); ni != object->mKdPvs.mEntries.end(); ni++) 
				{
					BspNode *node = (*ni).first;
	    
					// $$ JB TEMPORARY solution -> should add object PVS or explictly computed
					// BSP tree PVS
	    			if (leaf->mBspPvs.AddNodeSample(node))
						totalPvsSize++;
				}*/
			}
		} else 
		{
			// traverse tree
			BspInterior *interior = (BspInterior *)node;
      
			nodeStack.push(interior->GetFront());
			nodeStack.push(interior->GetBack());
		}
	}

	return totalPvsSize;
}

AxisAlignedBox3 BspTree::GetBoundingBox() const
{
	return mBox;
}

BspNode *BspTree::GetRoot() const
{
	return mRoot;
}

bool BspTree::StorePolys() const
{
	return mStorePolys;
}

void BspTree::EvaluateLeafStats(const BspTraversalData &data)
{
	// the node became a leaf -> evaluate stats for leafs
	BspLeaf *leaf = dynamic_cast<BspLeaf *>(data.mNode);

	if (data.mDepth >= mTermMaxDepth)
	{
		++ mStat.maxDepthNodes;
	}

	// store maximal and minimal depth
	if (data.mDepth > mStat.maxDepth)
		mStat.maxDepth = data.mDepth;
	if (data.mDepth < mStat.minDepth)
		mStat.minDepth = data.mDepth;

	// accumulate depth to compute average
	mStat.accumDepth += data.mDepth;

	if (leaf->mViewCell)
		++ mStat.viewCellLeaves;

#ifdef _DEBUG
	Debug << "BSP Traversal data. Depth: " << data.mDepth << " (max: " << mTermMaxDepth << "), #polygons: " << 
	  data.mPolygons->size() << " (max: " << mTermMaxPolygons << ")" << endl;
#endif
}

int BspTree::CastRay(Ray &ray)
{
	int hits = 0;
  
	stack<BspRayTraversalData> tStack;
  
	float maxt = 1e6;
	float mint = 0;

	Intersectable::NewMail();

	// test with tree bounding box
	if (!mBox.GetMinMaxT(ray, &mint, &maxt))
		return 0;

	if (mint < 0)
		mint = 0;
  
	maxt += Limits::Threshold;
  
	Vector3 entp = ray.Extrap(mint);
	Vector3 extp = ray.Extrap(maxt);
  
	BspNode *node = mRoot;
	BspNode *farChild;
	
	while (1)
	{
		if (!node->IsLeaf()) 
		{
			BspInterior *in = (BspInterior *) node;
			
			Plane3 *splitPlane = in->GetPlane();

			float t = 0;
			bool coplanar = false;
		
			int entSide = splitPlane->Side(entp);
			int extSide = splitPlane->Side(extp);

			Vector3 intersection;

			if (entSide < 0)
			{
				node = in->GetBack();

				if(extSide <= 0) 
					continue;
					
				farChild = in->GetFront(); // plane splits ray

			} else if (entSide > 0)
			{
				node = in->GetFront();

				if (extSide >= 0)
					continue;

				farChild = in->GetBack();	 // plane splits ray			
			}
			else // ray and plane are coincident // WHAT TO DO IN THIS CASE ?
			{
				node = in->GetFront();
				continue;
			}

			//-- split

			// push data for far child
			tStack.push(BspRayTraversalData(farChild, extp, maxt));

			// find intersection with plane between ray origin and exit point
			extp = splitPlane->FindIntersection(ray.GetLoc(), extp, &maxt);

		} else // compute intersections with objects in leaf
		{
			BspLeaf *leaf = dynamic_cast<BspLeaf *>(node);
			//ray.leaves.push_back(leaf);
      // TODO
   			/*ObjectContainer::const_iterator mi;
			for (mi = leaf->mObjects.begin(); mi != leaf->mObjects.end(); ++mi) 
			{
				Intersectable *object = *mi;
				if (!object->Mailed()) 
				{
					object->Mail();
					//ray.meshes.push_back(mesh);
					hits += object->CastRay(ray);
				}
			}*/

			if (hits && ray.GetType() == Ray::LOCAL_RAY)
				if (ray.intersections[0].mT <= maxt)
					break;
      
			// get the next node from the stack
			if (tStack.empty())
				break;
      
			entp = extp;
			mint = maxt;
			BspRayTraversalData &s  = tStack.top();

			node = s.mNode;
			extp = s.mExitPoint;
			maxt = s.mMaxT;

			tStack.pop();
		}
	}

	return hits;
}

bool BspTree::Export(const string filename)
{
	Exporter *exporter = Exporter::GetExporter(filename);

	if (exporter) 
	{
		exporter->ExportBspTree(*this);
		delete exporter;

		return true;
	}	

	return false;
}
//} // GtpVisibilityPreprocessor
